@@ -1,4 +1,4 @@
-# Copyright (c) 2016-2021 by Ron Frederick <ronf@timeheart.net> and others.
+# Copyright (c) 2016-2022 by Ron Frederick <ronf@timeheart.net> and others.
 #
 # This program and the accompanying materials are made available under
 # the terms of the Eclipse Public License v2.0 which accompanies this
@@ -23,8 +23,23 @@
 import re
 
 from functools import partial
+from typing import TYPE_CHECKING, Callable, Dict, List
+from typing import Optional, Set, Tuple, Union, cast
 from unicodedata import east_asian_width
 
+from .session import DataType
+
+
+if TYPE_CHECKING:
+    # pylint: disable=cyclic-import
+    from .channel import SSHServerChannel
+    from .session import SSHServerSession
+
+
+_CharDict = Dict[str, object]
+_CharHandler = Callable[['SSHLineEditor'], None]
+_KeyHandler = Callable[[str, int], Union[bool, Tuple[str, int]]]
+
 
 _DEFAULT_WIDTH = 80
 
@@ -34,7 +49,7 @@ _ansi_terminals = ('ansi', 'cygwin', 'linux', 'putty', 'screen', 'teraterm',
                    'rxvt-color')
 
 
-def _is_wide(ch):
+def _is_wide(ch: str) -> bool:
     """Return display width of character"""
 
     return east_asian_width(ch) in 'WF'
@@ -43,10 +58,14 @@ def _is_wide(ch):
 class SSHLineEditor:
     """Input line editor"""
 
-    def __init__(self, chan, session, history_size, max_line_length,
-                 term_type, width):
+    def __init__(self, chan: 'SSHServerChannel[str]',
+                 session: 'SSHServerSession[str]', line_echo: bool,
+                 history_size: int, max_line_length: int, term_type: str,
+                 width: int):
         self._chan = chan
         self._session = session
+        self._line_echo = line_echo
+        self._line_pending = False
         self._history_size = history_size if history_size > 0 else 0
         self._max_line_length = max_line_length
         self._wrap = term_type in _ansi_terminals
@@ -61,12 +80,12 @@ class SSHLineEditor:
         self._pos = 0
         self._line = ''
         self._bell_rung = False
-        self._early_wrap = set()
-        self._outbuf = []
-        self._keymap = {}
+        self._early_wrap: Set[int] = set()
+        self._outbuf: List[str] = []
+        self._keymap: _CharDict = {}
         self._key_state = self._keymap
         self._erased = ''
-        self._history = []
+        self._history: List[str] = []
         self._history_index = 0
 
         for func, keys in self._keylist:
@@ -75,7 +94,7 @@ class SSHLineEditor:
 
         self._build_printable()
 
-    def _add_key(self, key, func):
+    def _add_key(self, key: str, func: _CharHandler) -> None:
         """Add a key to the keymap"""
 
         keymap = self._keymap
@@ -84,11 +103,11 @@ class SSHLineEditor:
             if ch not in keymap:
                 keymap[ch] = {}
 
-            keymap = keymap[ch]
+            keymap = cast(_CharDict, keymap[ch])
 
         keymap[key[-1]] = func
 
-    def _del_key(self, key):
+    def _del_key(self, key: str) -> None:
         """Delete a key from the keymap"""
 
         keymap = self._keymap
@@ -97,25 +116,25 @@ class SSHLineEditor:
             if ch not in keymap:
                 return
 
-            keymap = keymap[ch]
+            keymap = cast(_CharDict, keymap[ch])
 
         keymap.pop(key[-1], None)
 
-    def _build_printable(self):
+    def _build_printable(self) -> None:
         """Build a regex of printable ASCII non-registered keys"""
 
-        def _escape(c):
+        def _escape(c: int) -> str:
             """Backslash escape special characters in regex character range"""
 
             ch = chr(c)
             return ('\\' if (ch in '-&|[]\\^~') else '') + ch
 
-        def _is_printable(ch):
+        def _is_printable(ch: str) -> bool:
             """Return if character is printable and has no handler"""
 
             return ch.isprintable() and ch not in keys
 
-        pat = []
+        pat: List[str] = []
         keys = self._keymap.keys()
         start = ord(' ')
         limit = 0x10000
@@ -138,14 +157,16 @@ class SSHLineEditor:
 
         self._printable = re.compile('[' + ''.join(pat) + ']*')
 
-    def _char_width(self, pos):
+    def _char_width(self, pos: int) -> int:
         """Return width of character at specified position"""
 
         return 1 + _is_wide(self._line[pos]) + ((pos + 1) in self._early_wrap)
 
-    def _determine_column(self, data, column, pos=None):
+    def _determine_column(self, data: str, column: int,
+                          pos: Optional[int] = None) -> Tuple[str, int]:
         """Determine new output column after output occurs"""
 
+        escaped = False
         offset = pos
         last_wrap_pos = pos
         wrapped_data = []
@@ -153,11 +174,19 @@ class SSHLineEditor:
         for ch in data:
             if ch == '\b':
                 column -= 1
+            elif ch == '\x1b':
+                escaped = True
+            elif escaped:
+                if ch == 'm':
+                    escaped = False
             else:
                 if _is_wide(ch) and (column % self._width) == self._width - 1:
                     column += 1
 
                     if pos is not None:
+                        assert last_wrap_pos is not None
+                        assert offset is not None
+
                         wrapped_data.append(data[last_wrap_pos - offset:
                                                  pos - offset])
                         last_wrap_pos = pos
@@ -173,12 +202,15 @@ class SSHLineEditor:
                 pos += 1
 
         if pos is not None:
-            wrapped_data.append(data[last_wrap_pos-offset:])
+            assert last_wrap_pos is not None
+            assert offset is not None
+
+            wrapped_data.append(data[last_wrap_pos - offset:])
             return ' '.join(wrapped_data), column
         else:
             return data, column
 
-    def _output(self, data, pos=None):
+    def _output(self, data: str, pos: Optional[int] = None) -> None:
         """Generate output and calculate new output column"""
 
         idx = data.rfind('\n')
@@ -197,14 +229,14 @@ class SSHLineEditor:
         if self._cursor and self._cursor % self._width == 0:
             self._outbuf.append(' \b')
 
-    def _ring_bell(self):
+    def _ring_bell(self) -> None:
         """Ring the terminal bell"""
 
         if not self._bell_rung:
             self._outbuf.append('\a')
             self._bell_rung = True
 
-    def _update_input_window(self, new_pos):
+    def _update_input_window(self, new_pos: int) -> int:
         """Update visible input window when not wrapping onto multiple lines"""
 
         line_len = len(self._line)
@@ -249,7 +281,7 @@ class SSHLineEditor:
 
         return column
 
-    def _move_cursor(self, column):
+    def _move_cursor(self, column: int) -> None:
         """Move the cursor to selected position in input line"""
 
         start_row = self._cursor // self._width
@@ -270,7 +302,7 @@ class SSHLineEditor:
 
         self._cursor = column
 
-    def _move_back(self, column):
+    def _move_back(self, column: int) -> None:
         """Move the cursor backward to selected position in input line"""
 
         if self._wrap:
@@ -279,7 +311,7 @@ class SSHLineEditor:
             self._outbuf.append('\b' * (self._cursor - column))
             self._cursor = column
 
-    def _clear_to_end(self):
+    def _clear_to_end(self) -> None:
         """Clear any remaining characters from previous input line"""
 
         column = self._cursor
@@ -295,14 +327,14 @@ class SSHLineEditor:
         self._move_back(column)
         self._end_column = column
 
-    def _erase_input(self):
+    def _erase_input(self) -> None:
         """Erase current input line"""
 
         self._move_cursor(self._start_column)
         self._clear_to_end()
         self._early_wrap.clear()
 
-    def _draw_input(self):
+    def _draw_input(self) -> None:
         """Draw current input line"""
 
         if self._line and self._echo:
@@ -319,7 +351,7 @@ class SSHLineEditor:
             self._end_column = self._cursor
             self._move_back(column)
 
-    def _reposition(self, new_pos, new_column):
+    def _reposition(self, new_pos: int, new_column: int) -> None:
         """Reposition the cursor to selected position in input"""
 
         if self._echo:
@@ -330,7 +362,7 @@ class SSHLineEditor:
 
         self._pos = new_pos
 
-    def _update_input(self, pos, column, new_pos):
+    def _update_input(self, pos: int, column: int, new_pos: int) -> None:
         """Update selected portion of current input line"""
 
         if self._echo:
@@ -356,7 +388,25 @@ class SSHLineEditor:
 
         self._pos = new_pos
 
-    def _insert_printable(self, data):
+    def _reset_line(self) -> None:
+        """Reset input line to empty"""
+
+        self._line = ''
+        self._left_pos = 0
+        self._right_pos = 0
+        self._pos = 0
+        self._start_column = self._cursor
+        self._end_column = self._cursor
+
+    def _reset_pending(self) -> None:
+        """Reset a pending echoed line if any"""
+
+        if self._line_pending:
+            self._erase_input()
+            self._reset_line()
+            self._line_pending = False
+
+    def _insert_printable(self, data: str) -> None:
         """Insert data into the input line"""
 
         line_len = len(self._line)
@@ -375,37 +425,35 @@ class SSHLineEditor:
 
             self._update_input(pos, self._cursor, new_pos)
 
-    def _end_line(self):
+    def _end_line(self) -> None:
         """End the current input line and send it to the session"""
 
-        if (self._echo and not self._wrap and
-                (self._left_pos > 0 or self._right_pos < len(self._line))):
-            self._output('\b' * (self._cursor - self._start_column) +
-                         self._line)
+        line = self._line
+
+        need_wrap = (self._echo and not self._wrap and
+                     (self._left_pos > 0 or self._right_pos < len(line)))
+
+        if self._line_echo or need_wrap:
+            if need_wrap:
+                self._output('\b' * (self._cursor - self._start_column) + line)
+            else:
+                self._move_to_end()
+
+            self._output('\r\n')
+            self._reset_line()
         else:
             self._move_to_end()
+            self._line_pending = True
 
-        self._output('\r\n')
-
-        self._start_column = 0
-        self._end_column = 0
-        self._cursor = 0
-        self._left_pos = 0
-        self._right_pos = 0
-        self._pos = 0
-
-        if self._echo and self._history_size and self._line:
-            self._history.append(self._line)
+        if self._echo and self._history_size and line:
+            self._history.append(line)
             self._history = self._history[-self._history_size:]
 
         self._history_index = len(self._history)
 
-        data = self._line + '\n'
-        self._line = ''
+        self._session.data_received(line + '\n', None)
 
-        self._session.data_received(data, None)
-
-    def _eof_or_delete(self):
+    def _eof_or_delete(self) -> None:
         """Erase character to the right, or send EOF if input line is empty"""
 
         if not self._line:
@@ -413,7 +461,7 @@ class SSHLineEditor:
         else:
             self._erase_right()
 
-    def _erase_left(self):
+    def _erase_left(self) -> None:
         """Erase character to the left"""
 
         if self._pos > 0:
@@ -424,7 +472,7 @@ class SSHLineEditor:
         else:
             self._ring_bell()
 
-    def _erase_right(self):
+    def _erase_right(self) -> None:
         """Erase character to the right"""
 
         if self._pos < len(self._line):
@@ -434,14 +482,14 @@ class SSHLineEditor:
         else:
             self._ring_bell()
 
-    def _erase_line(self):
+    def _erase_line(self) -> None:
         """Erase entire input line"""
 
         self._erased = self._line
         self._line = ''
         self._update_input(0, self._start_column, 0)
 
-    def _erase_to_end(self):
+    def _erase_to_end(self) -> None:
         """Erase to end of input line"""
 
         pos = self._pos
@@ -449,7 +497,7 @@ class SSHLineEditor:
         self._line = self._line[:pos]
         self._update_input(pos, self._cursor, pos)
 
-    def _handle_key(self, key, handler):
+    def _handle_key(self, key: str, handler: _KeyHandler) -> None:
         """Call an external key handler"""
 
         result = handler(self._line, self._pos)
@@ -462,7 +510,7 @@ class SSHLineEditor:
         elif result is False:
             self._ring_bell()
         else:
-            line, new_pos = result
+            line, new_pos = cast(Tuple[str, int], result)
 
             if new_pos < 0:
                 self._session.signal_received(line)
@@ -470,7 +518,7 @@ class SSHLineEditor:
                 self._line = line
                 self._update_input(0, self._start_column, new_pos)
 
-    def _history_prev(self):
+    def _history_prev(self) -> None:
         """Replace input with previous line in history"""
 
         if self._history_index > 0:
@@ -480,7 +528,7 @@ class SSHLineEditor:
         else:
             self._ring_bell()
 
-    def _history_next(self):
+    def _history_next(self) -> None:
         """Replace input with next line in history"""
 
         if self._history_index < len(self._history):
@@ -495,7 +543,7 @@ class SSHLineEditor:
         else:
             self._ring_bell()
 
-    def _move_left(self):
+    def _move_left(self) -> None:
         """Move left in input line"""
 
         if self._pos > 0:
@@ -505,7 +553,7 @@ class SSHLineEditor:
         else:
             self._ring_bell()
 
-    def _move_right(self):
+    def _move_right(self) -> None:
         """Move right in input line"""
 
         if self._pos < len(self._line):
@@ -515,28 +563,28 @@ class SSHLineEditor:
         else:
             self._ring_bell()
 
-    def _move_to_start(self):
+    def _move_to_start(self) -> None:
         """Move to start of input line"""
 
         self._reposition(0, self._start_column)
 
-    def _move_to_end(self):
+    def _move_to_end(self) -> None:
         """Move to end of input line"""
 
         self._reposition(len(self._line), self._end_column)
 
-    def _redraw(self):
+    def _redraw(self) -> None:
         """Redraw input line"""
 
         self._erase_input()
         self._draw_input()
 
-    def _insert_erased(self):
+    def _insert_erased(self) -> None:
         """Insert previously erased input"""
 
         self._insert_printable(self._erased)
 
-    def _send_break(self):
+    def _send_break(self) -> None:
         """Send break to session"""
 
         self._session.break_received(0)
@@ -557,28 +605,32 @@ class SSHLineEditor:
                 (_insert_erased, ('\x19',)),
                 (_send_break,    ('\x03', '\x1b[33~')))
 
-    def register_key(self, key, handler):
+    def register_key(self, key: str, handler: _KeyHandler) -> None:
         """Register a handler to be called when a key is pressed"""
 
         self._add_key(key, partial(SSHLineEditor._handle_key,
                                    key=key, handler=handler))
         self._build_printable()
 
-    def unregister_key(self, key):
+    def unregister_key(self, key: str) -> None:
         """Remove the handler associated with a key"""
 
         self._del_key(key)
         self._build_printable()
 
-    def set_input(self, line, pos):
+    def set_input(self, line: str, pos: int) -> None:
         """Set input line and cursor position"""
 
+        self._reset_pending()
+
         self._line = line
         self._update_input(0, self._start_column, pos)
 
-    def set_line_mode(self, line_mode):
+    def set_line_mode(self, line_mode: bool) -> None:
         """Enable/disable input line editing"""
 
+        self._reset_pending()
+
         if self._line and not line_mode:
             data = self._line
             self._erase_input()
@@ -588,9 +640,11 @@ class SSHLineEditor:
 
         self._line_mode = line_mode
 
-    def set_echo(self, echo):
+    def set_echo(self, echo: bool) -> None:
         """Enable/disable echoing of input in line mode"""
 
+        self._reset_pending()
+
         if self._echo and not echo:
             self._erase_input()
             self._echo = False
@@ -598,9 +652,11 @@ class SSHLineEditor:
             self._echo = True
             self._draw_input()
 
-    def set_width(self, width):
+    def set_width(self, width: int) -> None:
         """Set terminal line width"""
 
+        self._reset_pending()
+
         self._width = width or _DEFAULT_WIDTH
 
         if self._wrap:
@@ -609,7 +665,7 @@ class SSHLineEditor:
 
         self._redraw()
 
-    def process_input(self, data, datatype):
+    def process_input(self, data: str, datatype: DataType) -> None:
         """Process input from channel"""
 
         if self._line_mode:
@@ -617,18 +673,26 @@ class SSHLineEditor:
             idx = 0
 
             while idx < data_len:
+                self._reset_pending()
+
                 ch = data[idx]
                 idx += 1
 
                 if ch in self._key_state:
-                    self._key_state = self._key_state[ch]
-                    if callable(self._key_state):
+                    key_state = self._key_state[ch]
+
+                    if callable(key_state):
                         try:
-                            self._key_state(self)
+                            cast(_CharHandler, key_state)(self)
                         finally:
                             self._key_state = self._keymap
+                    else:
+                        self._key_state = cast(_CharDict, key_state)
                 elif self._key_state == self._keymap and ch.isprintable():
-                    match = self._printable.match(data, idx - 1)[0]
+                    match = self._printable.match(data, idx - 1)
+
+                    assert match is not None
+                    match = match[0]
 
                     if match:
                         self._insert_printable(match)
@@ -645,9 +709,19 @@ class SSHLineEditor:
         else:
             self._session.data_received(data, datatype)
 
-    def process_output(self, data):
+    def process_output(self, data: str) -> None:
         """Process output to channel"""
 
+        if self._line_pending:
+            if data.startswith(self._line):
+                self._start_column = self._cursor
+                data = data[len(self._line):]
+            else:
+                self._erase_input()
+
+            self._reset_line()
+            self._line_pending = False
+
         data = data.replace('\n', '\r\n')
 
         self._erase_input()
@@ -677,29 +751,36 @@ class SSHLineEditorChannel:
 
     """
 
-    def __init__(self, orig_chan, orig_session, history_size, max_line_length):
+    def __init__(self, orig_chan: 'SSHServerChannel[str]',
+                 orig_session: 'SSHServerSession[str]', line_echo: bool,
+                 history_size: int, max_line_length: int):
         self._orig_chan = orig_chan
         self._orig_session = orig_session
+        self._line_echo = line_echo
         self._history_size = history_size
         self._max_line_length = max_line_length
-        self._editor = None
+        self._editor: Optional[SSHLineEditor] = None
 
-    def __getattr__(self, attr):
+    def __getattr__(self, attr: str):
         """Delegate most channel functions to original channel"""
 
         return getattr(self._orig_chan, attr)
 
-    def create_editor(self):
+    def create_editor(self) -> Optional[SSHLineEditor]:
         """Create input line editor if encoding and terminal type are set"""
 
-        if self._encoding and self._term_type:
+        encoding, _ = self._orig_chan.get_encoding()
+        term_type = self._orig_chan.get_terminal_type()
+        width = self._orig_chan.get_terminal_size()[0]
+
+        if encoding and term_type:
             self._editor = SSHLineEditor(
-                self._orig_chan, self._orig_session, self._history_size,
-                self._max_line_length, self._term_type, self._term_size[0])
+                self._orig_chan, self._orig_session, self._line_echo,
+                self._history_size, self._max_line_length, term_type, width)
 
         return self._editor
 
-    def register_key(self, key, handler):
+    def register_key(self, key: str, handler: _KeyHandler) -> None:
         """Register a handler to be called when a key is pressed
 
            This method registers a handler function which will be called
@@ -735,9 +816,10 @@ class SSHLineEditorChannel:
 
         """
 
+        assert self._editor is not None
         self._editor.register_key(key, handler)
 
-    def unregister_key(self, key):
+    def unregister_key(self, key: str) -> None:
         """Remove the handler associated with a key
 
            This method removes a handler function associated with
@@ -752,18 +834,20 @@ class SSHLineEditorChannel:
 
         """
 
+        assert self._editor is not None
         self._editor.unregister_key(key)
 
-    def clear_input(self):
+    def clear_input(self) -> None:
         """Clear input line
 
            This method clears the current input line.
 
         """
 
+        assert self._editor is not None
         self._editor.set_input('', 0)
 
-    def set_input(self, line, pos):
+    def set_input(self, line: str, pos: int) -> None:
         """Clear input line
 
            This method sets the current input line and cursor position.
@@ -777,9 +861,10 @@ class SSHLineEditorChannel:
 
         """
 
+        assert self._editor is not None
         self._editor.set_input(line, pos)
 
-    def set_line_mode(self, line_mode):
+    def set_line_mode(self, line_mode: bool) -> None:
         """Enable/disable input line editing
 
            This method enabled or disables input line editing. When set,
@@ -795,9 +880,10 @@ class SSHLineEditorChannel:
         self._orig_chan.logger.info('%s line editor',
                                     'Enabling' if line_mode else 'Disabling')
 
+        assert self._editor is not None
         self._editor.set_line_mode(line_mode)
 
-    def set_echo(self, echo):
+    def set_echo(self, echo: bool) -> None:
         """Enable/disable echoing of input in line mode
 
            This method enables or disables echoing of input data when
@@ -812,9 +898,10 @@ class SSHLineEditorChannel:
         self._orig_chan.logger.info('%s echo',
                                     'Enabling' if echo else 'Disabling')
 
+        assert self._editor is not None
         self._editor.set_echo(echo)
 
-    def write(self, data, datatype=None):
+    def write(self, data: str, datatype: DataType = None) -> None:
         """Process data written to the channel"""
 
         if self._editor and datatype is None:
@@ -826,23 +913,25 @@ class SSHLineEditorChannel:
 class SSHLineEditorSession:
     """Input line editor session wrapper"""
 
-    def __init__(self, chan, orig_session):
+    def __init__(self, chan: SSHLineEditorChannel,
+                 orig_session: 'SSHServerSession[str]'):
         self._chan = chan
         self._orig_session = orig_session
-        self._editor = None
+        self._editor: Optional[SSHLineEditor] = None
 
-    def __getattr__(self, attr):
+    def __getattr__(self, attr: str):
         """Delegate most channel functions to original session"""
 
         return getattr(self._orig_session, attr)
 
-    def session_started(self):
+    def session_started(self) -> None:
         """Start a session for this newly opened server channel"""
 
         self._editor = self._chan.create_editor()
         self._orig_session.session_started()
 
-    def terminal_size_changed(self, width, height, pixwidth, pixheight):
+    def terminal_size_changed(self, width: int, height: int,
+                              pixwidth: int, pixheight: int) -> None:
         """The terminal size has changed"""
 
         if self._editor:
@@ -851,7 +940,7 @@ class SSHLineEditorSession:
         self._orig_session.terminal_size_changed(width, height,
                                                  pixwidth, pixheight)
 
-    def data_received(self, data, datatype):
+    def data_received(self, data: str, datatype: DataType) -> None:
         """Process data received from the channel"""
 
         if self._editor:
@@ -859,7 +948,7 @@ class SSHLineEditorSession:
         else:
             self._orig_session.data_received(data, datatype)
 
-    def eof_received(self):
+    def eof_received(self) -> Optional[bool]:
         """Process EOF received from the channel"""
 
         if self._editor:
