@@ -509,12 +509,14 @@ class SSHStreamSession(Generic[AnyStr]):
 
         recv_buf = self._recv_buf[datatype]
         data: List[AnyStr] = []
+        break_read = False
 
         async with self._read_locks[datatype]:
             while True:
                 while recv_buf and n != 0:
                     if isinstance(recv_buf[0], Exception):
                         if data:
+                            break_read = True
                             break
                         else:
                             exc = cast(Exception, recv_buf.pop(0))
@@ -542,7 +544,8 @@ class SSHStreamSession(Generic[AnyStr]):
                     continue
 
                 if n == 0 or (n > 0 and data and not exact) or \
-                        (n < 0 and recv_buf) or self._eof_received:
+                        (n < 0 and recv_buf) or \
+                        self._eof_received or break_read:
                     break
 
                 await self._block_read(datatype)
